#include <QPrinter>
#include <QPrintDialog>
#include <QFileDialog>
#include <QPixmap>
#include <QPainter>
#include <QSettings>
#include <QPushButton>
#include <QDir>
#include <QLabel>
#include <QComboBox>
#include <QTimer>
#include <QMessageBox>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "QPMain.h"

QPMain::QPMain()
 : printer(NULL), printDialog(NULL), fileDialog(NULL), image(NULL),
   scale_factor(1.0), batch_mode(false) {

   ui.setupUi(this);
   scale_label = new QLabel(" Preview Scale: ");
   QStringList scales;
   scales << "200%" << "150%" << "100%" << "75%" << "50%";
   scale_combo = new QComboBox;
   scale_combo->addItems(scales);
   scale_combo->setCurrentIndex(2);
   ui.toolBar->addWidget(scale_label);
   ui.toolBar->addWidget(scale_combo);
   ui.statusbar->show();
   connect(ui.actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));
   connect(ui.actionPrintSetup, SIGNAL(triggered()), this, SLOT(printSetup()));
   connect(ui.actionPrint, SIGNAL(triggered()), this, SLOT(print()));
   connect(ui.actionSaveConfig, SIGNAL(triggered()), this, SLOT(saveConfig()));
   connect(ui.actionLoadConfig, SIGNAL(triggered()), this, SLOT(loadConfig()));
   connect(ui.actionLoadImage, SIGNAL(triggered()), this, SLOT(loadImage()));
   connect(ui.actionAbout, SIGNAL(triggered()), this, SLOT(doAbout()));
   connect(ui.actionHome_Page, SIGNAL(triggered()), this, SLOT(showHomePage()));
   connect(scale_combo, SIGNAL(currentIndexChanged(int)),
           this, SLOT(changeScale(int)));
   image = new QPixmap(":/images/big-logo.png");
   ui.imageLabel->setPixmap(*image);
   ui.imageLabel->setMinimumSize(QSize(1,1));
   SetActionSensitivity();
}

QPMain::~QPMain() {
}

bool
QPMain::LoadConfigFile(QString config_name) {
   bool rval = false;
   QString found_config;
   if (access(qPrintable(config_name), R_OK) == 0) {
      found_config = config_name;
   } else {
      QString temp = config_name;
      if (! temp.endsWith(".qp")) {
         temp.append(".qp");
         if (access(qPrintable(temp), R_OK) == 0) {
            found_config = temp;
         }
      }
   }
   if (found_config.isEmpty()) {
      QString kewpiedir(getenv("HOME"));
      if (! kewpiedir.isEmpty()) {
         if (! kewpiedir.endsWith("/")) {
            kewpiedir.append("/");
         }
         kewpiedir.append(".kewpie/");
         QString temp = kewpiedir;
         temp.append(config_name);
         if (access(qPrintable(temp), R_OK) == 0) {
            found_config = temp;
         } else {
            if (! temp.endsWith(".qp")) {
               temp.append(".qp");
               if (access(qPrintable(temp), R_OK) == 0) {
                  found_config = temp;
               }
            }
         }
      }
   }
   if (found_config.isEmpty()) {
      QString errors(config_name);
      errors.append(" does not seem to be a valid printer config.");
      ShowErrors(errors);
   } else {
      InternLoadConfig(found_config);
      rval = true;
   }
   return rval;
}

void
QPMain::InternLoadConfig(QString config_name) {
   if (printer == NULL) {
      printer = new QPrinter(QPrinter::HighResolution);
   }
   QSettings ps(config_name, QSettings::IniFormat);
   printer->setPrinterName(ps.value("printerName").toString());
   QSizeF qsf(ps.value("paperSizeWidth").toDouble(),
              ps.value("paperSizeHeight").toDouble());
   printer->setPaperSize(qsf, QPrinter::Point);
   printer->setPaperSource(
      (QPrinter::PaperSource)ps.value("paperSource").toInt());
   printer->setPrintRange(
      (QPrinter::PrintRange)ps.value("printRange").toInt());
   printer->setResolution(ps.value("resolution").toInt());
   printer->setCollateCopies(ps.value("collateCopies").toBool());
   printer->setColorMode((QPrinter::ColorMode)ps.value("colorMode").toInt());
   printer->setDoubleSidedPrinting(ps.value("doubleSidedPrinting").toBool());
   printer->setDuplex((QPrinter::DuplexMode)ps.value("duplex").toInt());
   printer->setFromTo(ps.value("fromPage").toInt(),
                      ps.value("toPage").toInt());
   printer->setFullPage(ps.value("fullPage").toBool());
   printer->setNumCopies(ps.value("numCopies").toInt());
   printer->setOrientation(
      (QPrinter::Orientation)ps.value("orientation").toInt());
   printer->setOutputFormat(
      (QPrinter::OutputFormat)ps.value("outputFormat").toInt());
   printer->setPageOrder(
      (QPrinter::PageOrder)ps.value("pageOrder").toInt());
   printer->setPaperSize(QSizeF(ps.value("paperSizeWidth").toDouble(),
                                ps.value("paperSizeHeight").toDouble()),
                         QPrinter::Point);
   printer->setPageMargins(ps.value("leftMargin").toDouble(),
                           ps.value("topMargin").toDouble(),
                           ps.value("rightMargin").toDouble(),
                           ps.value("bottomMargin").toDouble(),
                           QPrinter::Point);
   ui.imageLabel->clear();
   QTimer::singleShot(0, this, SLOT(doViewChange()));
}

bool
QPMain::LoadImageFile(QString image_name) {
   bool rval = false;
   if (! image_name.isEmpty()) {
      QPixmap * new_image = new QPixmap(image_name);
      if ((new_image->width() == 0) || (new_image->height() == 0)) {
         QString errors(image_name);
         errors.append(" does not seem to be a valid image.");
         delete new_image;
         new_image = NULL;
         ShowErrors(errors);
      }
      if (new_image != NULL) {
         if (image != NULL) {
            delete image;
         }
         image = new_image;
         rval = true;
         QTimer::singleShot(0, this, SLOT(doViewChange()));
      }
   }
   return rval;
}

void
QPMain::SetBatchMode() {
   batch_mode = true;
}

void
QPMain::BatchPrint() {
   print();
   QTimer::singleShot(0, qApp, SLOT(quit()));
}

void
QPMain::printSetup() {
   if (printer == NULL) {
      printer = new QPrinter(QPrinter::HighResolution);
      SetActionSensitivity();
   }
   if (printDialog == NULL) {
      printDialog = new QPrintDialog(printer, this);

      // Nasty hack to make Print button just say OK instead

      QList<QPushButton *> allPButtons =
         printDialog->findChildren<QPushButton *>();
      for (int i = 0; i < allPButtons.size(); ++i) {
         QPushButton * btn = allPButtons.at(i);
         QString btn_text = btn->text();
         if (btn_text == QString("&Print")) {
            btn->setText("O&K");
         }
      }
   }

   if (printDialog->exec() == QDialog::Accepted) {
      ui.imageLabel->clear();
      SetViewSize();
   }
}

void
QPMain::print()
{
   if (printer != NULL) {
      QPainter painter(printer);
      QRect rect = painter.viewport();
      QSize size = image->size();
      size.scale(rect.size(), Qt::IgnoreAspectRatio);
      painter.setViewport(rect.x(), rect.y(), size.width(), size.height());
      painter.setWindow(image->rect());
      painter.drawPixmap(0, 0, *image);
   }
}

void
QPMain::saveConfig() {
   if (printer == NULL) {
      return;
   }
   QFileDialog * fd = SaveConfigDialog();
   QString config_file;
   if (fd->exec() == QDialog::Accepted) {
      config_file = fd->selectedFiles().at(0);
   }
   if (! config_file.isEmpty()) {
      QSettings ps(config_file, QSettings::IniFormat);
      ps.clear();
      ps.setValue("printerName", printer->printerName());
      QSizeF qsf = printer->paperSize(QPrinter::Point);
      ps.setValue("paperSizeWidth", qsf.width());
      ps.setValue("paperSizeHeight", qsf.height());
      ps.setValue("paperSource", printer->paperSource());
      ps.setValue("printRange", printer->printRange());
      ps.setValue("resolution", printer->resolution());
      ps.setValue("collateCopies", printer->collateCopies());
      ps.setValue("colorMode", printer->colorMode());
      ps.setValue("doubleSidedPrinting", printer->doubleSidedPrinting());
      ps.setValue("duplex", printer->duplex());
      ps.setValue("fromPage", printer->fromPage());
      ps.setValue("toPage", printer->toPage());
      ps.setValue("fullPage", printer->fullPage());
      ps.setValue("numCopies", printer->numCopies());
      ps.setValue("orientation", printer->orientation());
      ps.setValue("outputFormat", printer->outputFormat());
      ps.setValue("pageOrder", printer->pageOrder());
      qsf = printer->paperSize(QPrinter::Point);
      ps.setValue("paperSizeWidth", qsf.width());
      ps.setValue("paperSizeHeight", qsf.height());
      QRectF qrf1 = printer->pageRect(QPrinter::Point);
      QRectF qrf2 = printer->paperRect(QPrinter::Point);
      qreal left_margin = qrf1.x() - qrf2.x();
      qreal top_margin = qrf1.y() - qrf2.y();
      qreal right_margin = qrf2.width() - (qrf1.width() + left_margin);
      qreal bottom_margin = qrf2.height() - (qrf1.height() + top_margin);
      ps.setValue("leftMargin", left_margin);
      ps.setValue("topMargin", top_margin);
      ps.setValue("rightMargin", right_margin);
      ps.setValue("bottomMargin", bottom_margin);
   }
}

void
QPMain::loadConfig() {
   QString config_file;
   QFileDialog * fd = LoadConfigDialog();
   if (fd->exec() == QDialog::Accepted) {
      config_file = fd->selectedFiles().at(0);
   }
   if (! config_file.isEmpty()) {
      InternLoadConfig(config_file);
   }
}

void
QPMain::loadImage() {
   QString image_file =
      QFileDialog::getOpenFileName(this, "Load Image",
                                   QDir::current().absolutePath(),
                                   "Images (*.png *.xpm *.jpg)");
   if (! image_file.isEmpty()) {
      LoadImageFile(image_file);
   }
}

void
QPMain::changeScale(int scale_index) {
   switch (scale_index) {
   case 0:              // "200%"
      scale_factor = 2.0;
      break;
   case 1:              // "150%"
      scale_factor = 1.5;
      break;
   case 2:              // "100%"
      scale_factor = 1.0;
      break;
   case 3:              // "75%"
      scale_factor = 0.75;
      break;
   case 4:              // "50%"
      scale_factor = 0.5;
      break;
   }
   SetViewSize();
}

void
QPMain::doViewChange() {
   if (! batch_mode) {
      SetViewSize();
      SetActionSensitivity();
   }
}

void
QPMain::SetActionSensitivity() {
   ui.actionLoadConfig->setEnabled(true);
   ui.actionSaveConfig->setEnabled(printer != NULL);
   ui.actionPrintSetup->setEnabled(true);
   ui.actionQuit->setEnabled(true);
   ui.actionLoadImage->setEnabled(printer != NULL);
   ui.actionPrint->setEnabled((printer != NULL) && (image != NULL));
}

void
QPMain::SetViewSize() {

   if (printer == NULL) {
      return;
   }

   QRectF paper = printer->paperRect(QPrinter::Inch);
   QRectF page = printer->pageRect(QPrinter::Inch);

   // First compute the pixel width and height I want the central
   // widget to have.

   int dpi_width = physicalDpiX();
   int dpi_height = physicalDpiY();
   int pix_width =
      (int)((((qreal)dpi_width) * paper.width())*scale_factor + 0.5);
   int pix_height =
      (int)((((qreal)dpi_height) * paper.height())*scale_factor + 0.5);
   int img_x =
      (int)((((qreal)dpi_width) * page.x())*scale_factor + 0.5);
   int img_y =
      (int)((((qreal)dpi_height) * page.y())*scale_factor + 0.5);
   int img_width =
      (int)((((qreal)dpi_width) * page.width())*scale_factor + 0.5);
   int img_height =
      (int)((((qreal)dpi_height) * page.height())*scale_factor + 0.5);

   int grow_width = pix_width - ui.imageLabel->width();
   int grow_height = pix_height - ui.imageLabel->height();

   ui.imageLabel->clear();
   ui.imageLabel->setFixedSize(QSize(pix_width, pix_height));
   ui.imageLabel->setMinimumSize(QSize(1,1));

   if (image != NULL) {
      QPixmap preview(pix_width, pix_height);
      QPixmap margins(":/images/margins.png");
      QPainter paint;
      paint.begin(&preview);
      paint.drawTiledPixmap(0, 0, pix_width, pix_height, margins);
      paint.fillRect(img_x, img_y, img_width, img_height, Qt::white);
      paint.drawPixmap(img_x, img_y,
                       image->scaled(img_width, img_height,
                                     Qt::IgnoreAspectRatio,
                                     Qt::SmoothTransformation));
      paint.end();
      ui.imageLabel->setPixmap(preview);
   }

   // Now find the current width and height of the central widget
   // and change the size of the main window by the difference in size
   // of the central widget (which should have the side effect of leaving
   // the central widget the size we want).

   resize(width() + grow_width, height() + grow_height);
}

QFileDialog *
QPMain::ConfigFileDialog() {
   if (fileDialog == NULL) {
      QString kewpiedir(getenv("HOME"));
      if (! kewpiedir.isEmpty()) {
         if (! kewpiedir.endsWith("/")) {
            kewpiedir.append("/");
         }
         kewpiedir.append(".kewpie");
         mkdir(qPrintable(kewpiedir), 0777);
      }
      fileDialog = new QFileDialog(this, QString(), kewpiedir, "Configs (*.qp)");
      fileDialog->setDefaultSuffix("qp");
      fileDialog->setOption(QFileDialog::DontResolveSymlinks);
   }
   return fileDialog;
}

QFileDialog *
QPMain::SaveConfigDialog() {
   QFileDialog * fd = ConfigFileDialog();
   fd->setAcceptMode(QFileDialog::AcceptSave);
   fd->setOption(QFileDialog::DontConfirmOverwrite, false);
   fd->setOption(QFileDialog::ReadOnly, false);
   fd->setFileMode(QFileDialog::AnyFile);
   fd->setWindowTitle("Save Printer Config");
   return fd;
}

QFileDialog *
QPMain::LoadConfigDialog() {
   QFileDialog * fd = ConfigFileDialog();
   fd->setAcceptMode(QFileDialog::AcceptOpen);
   fd->setOption(QFileDialog::DontConfirmOverwrite, true);
   fd->setOption(QFileDialog::ReadOnly, true);
   fd->setFileMode(QFileDialog::ExistingFile);
   fd->setWindowTitle("Load Printer Config");
   return fd;
}

void
QPMain::ShowErrors(QString msg) {
   if (batch_mode) {
      fputs(qPrintable(msg), stderr);
      fputc('\n', stderr);
   } else {
      if (! msg_accum.isEmpty()) {
         msg_accum.append("\n");
      }
      msg_accum.append(msg);
      QTimer::singleShot(0, this, SLOT(doShowErrors()));
   }
}

void
QPMain::doShowErrors() {
   QString msg = msg_accum;
   msg_accum.clear();
   QMessageBox::critical(this, "Kewpie Error", msg);
}

void
QPMain::doAbout() {
   QString msg("Kewpie 1.0 Image Print Utility");
   if (batch_mode) {
      fputs(qPrintable(msg), stdout);
      fputc('\n', stdout);
   } else {
      QMessageBox::about(this, "Kewpie Version", msg);
   }
}

void
QPMain::showHomePage() {

   // Use double fork gimmick to run xdg-open passing it the
   // url for kewpie home page. (As near as I can tell, xdg-open is the
   // best I can do at running the "preferred" browser for a user).

   pid_t first_kid = fork();
   if (first_kid == 0) {
      // I am the first kid, all I do is fork again and
      // exit after redirecting I/O to /dev/null
      for (int i = 3; i < 32768; ++i) {
         ::close(i);
      }
      int fd = ::open("/dev/null", O_RDONLY);
      ::dup2(fd, fileno(stdin));
      ::close(fd);
      fd = ::open("/dev/null", O_WRONLY);
      ::dup2(fd, fileno(stdout));
      ::dup2(fd, fileno(stderr));
      ::close(fd);
      pid_t grandkid = fork();
      if (grandkid == 0) {

         // OK, here is where I run xdg-open to show the kewpie
         // home page. I'll be owned by init so there won't be any
         // defunct processes around to clutter things up.

         execlp("xdg-open", "xdg-open",
            "http://home.att.net/~Tom.Horsley/kewpie/kewpie.html", NULL);
      }
      _exit(0);
   } else if (first_kid != (pid_t)-1) {

      // I am parent, wait on first kid which should not block event loop
      // since he is gonna exit right away.

      int dummy;
      waitpid(first_kid, &dummy, 0);
   }
}
