#include <QApplication>
#include <QString>
#include <QTimer>
#include <stdio.h>
#include <string.h>
#include "QPMain.h"

int
main(int argc, char **argv)
{
   QString init_config_file;
   QString init_image_file;
   bool batchmode = false;
   bool config_loaded = false;
   bool image_loaded = false;
   bool print_version = false;
   bool print_help = false;
   int rval = 0;
   QApplication app(argc, argv);
   for (int i = 1; i < argc; ++i) {
      if ((strcmp(argv[i], "-config") == 0) && (i < argc - 1)) {
         init_config_file.append(argv[i+1]);
         ++i;
      } else if ((strcmp(argv[i], "-image") == 0) && (i < argc - 1)) {
         init_image_file.append(argv[i+1]);
         ++i;
      } else if (strcmp(argv[i], "-batch") == 0) {
         batchmode = true;
      } else if ((strcmp(argv[i], "-version") == 0) ||
                 (strcmp(argv[i], "-v") == 0)) {
         batchmode = true;
         print_version = true;
      } else if ((strcmp(argv[i], "-help") == 0) ||
                 (strcmp(argv[i], "-h") == 0)) {
         batchmode = true;
         print_help = true;
      } else {
         fprintf(stderr, "kewpie: Unrecognized argument: %s\n", argv[i]);
         batchmode = true;
         print_help = true;
         rval = 2;
      }
   }
   QPMain mainWin;
   if (batchmode) {
      mainWin.SetBatchMode();
   }
   if (print_version) {
      mainWin.doAbout();
      return rval;
   }
   if (print_help) {
      fputs(
"usage: kewpie [-config file] [-image file] [-batch] [-version] [-help]\n\
-config  Name of the printer config to load\n\
-image   Name of the image to print\n\
-batch   Operate in batch mode, no GUI windows\n\
-version Print version and exit\n\
-help    Print this help and exit\n\
\n\
kewpie is a utility for printing images with precise margins. Primarily\n\
designed for printing DVDs and CDs, but not restricted to that.\n\
Home page: http://home.att.net/~Tom.Horsley/kewpie/kewpie.html\n", stdout);
      return rval;
   }
   if (! init_config_file.isEmpty()) {
      config_loaded = mainWin.LoadConfigFile(init_config_file);
   }
   if (config_loaded) {
      if (! init_image_file.isEmpty()) {
         image_loaded = mainWin.LoadImageFile(init_image_file);
      }
   }
   if (batchmode && ((! config_loaded) || (! image_loaded))) {
      return 2;
   }
   if (! batchmode) {
      mainWin.resize(610, 706);
      mainWin.show();
   } else {
      QTimer::singleShot(0, &mainWin, SLOT(BatchPrint()));
   }
   return app.exec();
}
